// Copyright 2021 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0
use crate::Context;

impl Context {
    // Missing function: HMAC_Start
    // Missing function: MAC_Start
    // Missing function: HashSequenceStart
    // Missing function: SequenceUpdate
    // Missing function: SequenceComplete
    // Missing function: EventSequenceComplete
}
