// Copyright 2021 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0
use crate::Context;

impl Context {
    // Missing function: ReadClock
    // Missing function: ClockSet
    // Missing function: ClockRateAdjust
}
