// Copyright 2020 Contributors to the Parsec project.
// SPDX-License-Identifier: Apache-2.0

/// This module contains code that deaals with non volatile storage
/// in the TPM.
///
mod public;

pub use public::{NvPublic, NvPublicBuilder};
